import numpy as np
import pandas as pd

import spotify_confidence


class TestBootstrap(object):
    def setup(self):
        np.random.seed(123)
        n_bootstraps = int(5e5)
        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test3"],
                "bootstraps": [
                    np.random.randn(n_bootstraps) + 0.5,
                    np.random.randn(n_bootstraps) + 0.4,
                    np.random.randn(n_bootstraps) + 0.2,
                    np.random.randn(n_bootstraps) + 1 / 3,
                ],
                "method": ["bootstrap", "bootstrap", "bootstrap", "bootstrap"],
            }
        )

        self.test = spotify_confidence.Experiment(
            self.data,
            numerator_column=None,
            numerator_sum_squares_column=None,
            denominator_column=None,
            categorical_group_columns="variation_name",
            method_column="method",
            bootstrap_samples_column="bootstraps",
        )

    def test_summary(self):
        summary_df = self.test.summary()
        assert len(summary_df) == len(self.data)

    def test_difference(self):
        difference_df = self.test.difference(level_1="control", level_2="test", absolute=True)
        assert len(difference_df) == 1
        assert np.allclose(difference_df["difference"][0], 0.1, atol=1e-2)

    def test_difference_absolute_false(self):
        difference_df = self.test.difference(level_1="control", level_2="test", absolute=False)
        assert len(difference_df) == 1
        assert np.allclose(difference_df["difference"][0], 0.1 / 0.4, atol=1e-2)

    def test_multiple_difference(self):
        difference_df = self.test.multiple_difference(level="control", level_as_reference=True)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"][0], 0.1, atol=1e-2)

    def test_multiple_difference_level_as_reference_false(self):
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=False, absolute=True)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"], -difference_df_true_true["difference"], atol=0)
        assert np.allclose(difference_df["ci_lower"], -difference_df_true_true["ci_upper"], atol=0)
        assert np.allclose(difference_df["ci_upper"], -difference_df_true_true["ci_lower"], atol=0)
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_multiple_difference_absolute_false(self):
        control_mean = self.test.summary().query("variation_name == 'control'")["point_estimate"].values[0]
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=True, absolute=False)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"], difference_df_true_true["difference"] / control_mean, atol=0)
        assert np.allclose(difference_df["ci_lower"], difference_df_true_true["ci_lower"] / control_mean, atol=0)
        assert np.allclose(difference_df["ci_upper"], difference_df_true_true["ci_upper"] / control_mean, atol=0)
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_multiple_difference_level_as_reference_false_absolute_false(self):
        reference_mean = self.test.summary().query("variation_name != 'control'")["point_estimate"]
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=False, absolute=False)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(
            difference_df["difference"], -difference_df_true_true["difference"] / reference_mean.values, atol=0
        )
        assert np.allclose(
            difference_df["ci_lower"], -difference_df_true_true["ci_upper"] / reference_mean.values, atol=0
        )
        assert np.allclose(
            difference_df["ci_upper"], -difference_df_true_true["ci_lower"] / reference_mean.values, atol=0
        )
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_summary_plot(self):
        chart_grid = self.test.summary_plot()
        assert len(chart_grid.charts) == 1

    def test_difference_plot(self):
        chartgrid = self.test.difference_plot(level_1="control", level_2="test")
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot(self):
        chartgrid = self.test.multiple_difference_plot(level="control", level_as_reference=True)
        assert len(chartgrid.charts) == 1



from typing import Tuple

import numpy as np
from pandas import DataFrame, Series

from spotify_confidence.analysis.constants import CI_LOWER, CI_UPPER, SFX1, SFX2


class BootstrapComputer(object):
    def __init__(self, bootstrap_samples_column, interval_size):
        self._bootstrap_samples = bootstrap_samples_column
        self._interval_size = interval_size

    def _point_estimate(self, row: Series) -> float:
        return row[self._bootstrap_samples].mean()

    def _variance(self, row: Series) -> float:
        variance = row[self._bootstrap_samples].var()

        if variance < 0:
            raise ValueError("Computed variance is negative. " "Please check your inputs.")
        return variance

    def _std_err(self, row: Series) -> float:
        return None

    def _add_point_estimate_ci(self, row: DataFrame) -> Series:

        row[CI_LOWER] = np.percentile(row[self._bootstrap_samples], 100 * (1 - self._interval_size) / 2)
        row[CI_UPPER] = np.percentile(row[self._bootstrap_samples], 100 * (1 - (1 - self._interval_size) / 2))
        return row

    def _p_value(self, row) -> float:
        return -1

    def _ci(self, row, alpha_column: str) -> Tuple[float, float]:
        differences = row[self._bootstrap_samples + SFX2] - row[self._bootstrap_samples + SFX1]
        lower = np.percentile(differences, 100 * row[alpha_column] / 2)
        upper = np.percentile(differences, 100 * (1 - row[alpha_column] / 2))
        return lower, upper

    def _achieved_power(self, df: DataFrame, mde: float, alpha: float) -> DataFrame:
        return None



from typing import Tuple

import numpy as np
from pandas import DataFrame, Series
from statsmodels.stats.weightstats import _tconfint_generic, _tstat_generic

from spotify_confidence.analysis.confidence_utils import power_calculation
from spotify_confidence.analysis.constants import (
    POINT_ESTIMATE,
    CI_LOWER,
    CI_UPPER,
    VARIANCE,
    TWO_SIDED,
    SFX1,
    SFX2,
    STD_ERR,
    PREFERENCE_TEST,
    NULL_HYPOTHESIS,
    DIFFERENCE,
)


class TTestComputer(object):
    def __init__(self, numerator, numerator_sumsq, denominator, ordinal_group_column, interval_size):
        self._numerator = numerator
        self._numerator_sumsq = numerator_sumsq
        self._denominator = denominator
        self._ordinal_group_column = ordinal_group_column
        self._interval_size = interval_size

    def _point_estimate(self, row: Series) -> float:
        if row[self._denominator] == 0:
            raise ValueError("""Can't compute point estimate: denominator is 0""")
        return row[self._numerator] / row[self._denominator]

    def _variance(self, row: Series) -> float:
        variance = row[self._numerator_sumsq] / row[self._denominator] - row[POINT_ESTIMATE] ** 2
        if variance < 0:
            raise ValueError("Computed variance is negative. " "Please check your inputs.")
        return variance

    def _std_err(self, row: Series) -> float:
        return np.sqrt(
            row[VARIANCE + SFX1] / row[self._denominator + SFX1] + row[VARIANCE + SFX2] / row[self._denominator + SFX2]
        )

    def _add_point_estimate_ci(self, row: DataFrame) -> Series:
        row[CI_LOWER], row[CI_UPPER] = _tconfint_generic(
            mean=row[POINT_ESTIMATE],
            std_mean=np.sqrt(row[VARIANCE] / row[self._denominator]),
            dof=row[self._denominator] - 1,
            alpha=1 - self._interval_size,
            alternative=TWO_SIDED,
        )
        return row

    def _dof(self, row: Series) -> float:
        v1, v2 = row[VARIANCE + SFX1], row[VARIANCE + SFX2]
        n1, n2 = row[self._denominator + SFX1], row[self._denominator + SFX2]
        return (v1 / n1 + v2 / n2) ** 2 / ((v1 / n1) ** 2 / (n1 - 1) + (v2 / n2) ** 2 / (n2 - 1))

    def _p_value(self, row: Series) -> float:
        _, p_value = _tstat_generic(
            value1=row[POINT_ESTIMATE + SFX2],
            value2=row[POINT_ESTIMATE + SFX1],
            std_diff=row[STD_ERR],
            dof=self._dof(row),
            alternative=row[PREFERENCE_TEST],
            diff=row[NULL_HYPOTHESIS],
        )
        return p_value

    def _ci(self, row: Series, alpha_column: str) -> Tuple[float, float]:
        return _tconfint_generic(
            mean=row[DIFFERENCE],
            std_mean=row[STD_ERR],
            dof=self._dof(row),
            alpha=row[alpha_column],
            alternative=row[PREFERENCE_TEST],
        )

    def _achieved_power(self, df: DataFrame, mde: float, alpha: float) -> DataFrame:
        v1, v2 = df[VARIANCE + SFX1], df[VARIANCE + SFX2]
        n1, n2 = df[self._denominator + SFX1], df[self._denominator + SFX2]

        var_pooled = ((n1 - 1) * v1 + (n2 - 1) * v2) / (n1 + n2 - 2)

        return power_calculation(mde, var_pooled, alpha, n1, n2)



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Iterable, List, Tuple
from warnings import warn

import numpy as np
from pandas import DataFrame, Series, concat
from statsmodels.stats.multitest import multipletests

from spotify_confidence.analysis.abstract_base_classes.confidence_computer_abc import ConfidenceComputerABC
from spotify_confidence.analysis.confidence_utils import (
    get_remaning_groups,
    validate_levels,
    level2str,
    listify,
    add_nim_columns,
    validate_and_rename_columns,
    add_mde_columns,
    get_all_categorical_group_columns,
    get_all_group_columns,
    validate_data,
    remove_group_columns,
)
from spotify_confidence.analysis.constants import (
    POINT_ESTIMATE,
    VARIANCE,
    CI_LOWER,
    CI_UPPER,
    DIFFERENCE,
    P_VALUE,
    SFX1,
    SFX2,
    STD_ERR,
    ALPHA,
    ADJUSTED_ALPHA,
    ADJUSTED_ALPHA_POWER_SAMPLE_SIZE,
    POWER,
    POWERED_EFFECT,
    ADJUSTED_POWER,
    ADJUSTED_P,
    ADJUSTED_LOWER,
    ADJUSTED_UPPER,
    IS_SIGNIFICANT,
    REQUIRED_SAMPLE_SIZE,
    NULL_HYPOTHESIS,
    ALTERNATIVE_HYPOTHESIS,
    NIM,
    PREFERENCE,
    PREFERENCE_TEST,
    TWO_SIDED,
    PREFERENCE_DICT,
    BONFERRONI,
    HOLM,
    HOMMEL,
    SIMES_HOCHBERG,
    SIDAK,
    HOLM_SIDAK,
    FDR_BH,
    FDR_BY,
    FDR_TSBH,
    FDR_TSBKY,
    SPOT_1_HOLM,
    SPOT_1_HOMMEL,
    SPOT_1_SIMES_HOCHBERG,
    SPOT_1_SIDAK,
    SPOT_1_HOLM_SIDAK,
    SPOT_1_FDR_BH,
    SPOT_1_FDR_BY,
    SPOT_1_FDR_TSBH,
    SPOT_1_FDR_TSBKY,
    BONFERRONI_ONLY_COUNT_TWOSIDED,
    BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
    SPOT_1,
    CORRECTION_METHODS,
    BOOTSTRAP,
    CHI2,
    TTEST,
    ZTEST,
    NIM_TYPE,
    CORRECTION_METHODS_THAT_REQUIRE_METRIC_INFO,
)
from spotify_confidence.analysis.frequentist.confidence_computers.bootstrap_computer import BootstrapComputer
from spotify_confidence.analysis.frequentist.confidence_computers.chi_squared_computer import ChiSquaredComputer
from spotify_confidence.analysis.frequentist.confidence_computers.t_test_computer import TTestComputer
from spotify_confidence.analysis.frequentist.confidence_computers.z_test_computer import ZTestComputer
from spotify_confidence.analysis.frequentist.sequential_bound_solver import bounds


def sequential_bounds(t: np.array, alpha: float, sides: int):
    return bounds(t, alpha, rho=2, ztrun=8, sides=sides, max_nints=1000)


class GenericComputer(ConfidenceComputerABC):
    def __init__(
        self,
        data_frame: DataFrame,
        numerator_column: str,
        numerator_sum_squares_column: str,
        denominator_column: str,
        categorical_group_columns: Union[str, Iterable],
        ordinal_group_column: str,
        interval_size: float,
        correction_method: str,
        method_column: str,
        bootstrap_samples_column: str,
        metric_column: Union[str, None],
        treatment_column: Union[str, None],
        power: float,
    ):

        self._df = data_frame
        self._numerator = numerator_column
        self._numerator_sumsq = numerator_sum_squares_column
        if self._numerator is not None and (self._numerator_sumsq is None or self._numerator_sumsq == self._numerator):
            if (data_frame[numerator_column] <= data_frame[denominator_column]).all():
                # Treat as binomial data
                self._numerator_sumsq = self._numerator
            else:
                raise ValueError(
                    f"numerator_sum_squares_column missing or same as "
                    f"numerator_column, but since {numerator_column} is not "
                    f"always smaller than {denominator_column} it can't be "
                    f"binomial data. Please check your data."
                )

        self._denominator = denominator_column
        self._categorical_group_columns = get_all_categorical_group_columns(
            categorical_group_columns, metric_column, treatment_column
        )
        self._segments = remove_group_columns(self._categorical_group_columns, metric_column)
        self._segments = remove_group_columns(self._segments, treatment_column)
        self._ordinal_group_column = ordinal_group_column
        self._metric_column = metric_column
        self._interval_size = interval_size
        self._power = power
        self._treatment_column = treatment_column

        if correction_method.lower() not in CORRECTION_METHODS:
            raise ValueError(f"Use one of the correction methods " + f"in {CORRECTION_METHODS}")
        self._correction_method = correction_method
        self._method_column = method_column

        self._single_metric = False
        if self._metric_column is not None and data_frame.groupby(self._metric_column).ngroups == 1:
            self._single_metric = True
            self._categorical_group_columns = remove_group_columns(
                self._categorical_group_columns, self._metric_column
            )

        self._all_group_columns = get_all_group_columns(self._categorical_group_columns, self._ordinal_group_column)

        self._bootstrap_samples_column = bootstrap_samples_column

        columns_that_must_exist = []
        if (
            CHI2 in self._df[self._method_column]
            or TTEST in self._df[self._method_column]
            or ZTEST in self._df[self._method_column]
        ):
            columns_that_must_exist += [self._numerator, self._denominator]
            columns_that_must_exist += [] if self._numerator_sumsq is None else [self._numerator_sumsq]
        if BOOTSTRAP in self._df[self._method_column]:
            columns_that_must_exist += [self._bootstrap_samples_column]

        validate_data(self._df, columns_that_must_exist, self._all_group_columns, self._ordinal_group_column)

        self._sufficient = None
        self._computers = {
            CHI2: ChiSquaredComputer(
                self._numerator,
                self._numerator_sumsq,
                self._denominator,
                self._ordinal_group_column,
                self._interval_size,
            ),
            TTEST: TTestComputer(
                self._numerator,
                self._numerator_sumsq,
                self._denominator,
                self._ordinal_group_column,
                self._interval_size,
            ),
            ZTEST: ZTestComputer(
                self._numerator,
                self._numerator_sumsq,
                self._denominator,
                self._ordinal_group_column,
                self._interval_size,
            ),
            BOOTSTRAP: BootstrapComputer(self._bootstrap_samples_column, self._interval_size),
        }

    @property
    def _confidence_computers(self):
        return self._computers

    def compute_summary(self, verbose: bool) -> DataFrame:
        return (
            self._sufficient_statistics
            if verbose
            else self._sufficient_statistics[
                self._all_group_columns
                + ([self._metric_column] if self._metric_column is not None and self._single_metric else [])
                + [c for c in [self._numerator, self._denominator] if c is not None]
                + [POINT_ESTIMATE, CI_LOWER, CI_UPPER]
            ]
        )

    @property
    def _sufficient_statistics(self) -> DataFrame:
        if self._sufficient is None:
            self._sufficient = (
                self._df.assign(**{POINT_ESTIMATE: self._point_estimate})
                .assign(**{VARIANCE: self._variance})
                .pipe(self._add_point_estimate_ci)
            )
        return self._sufficient

    def compute_difference(
        self,
        level_1: Union[str, Iterable],
        level_2: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        difference_df = self._compute_differences(
            level_columns=level_columns,
            levels=[(level_1, level_2)],
            absolute=absolute,
            groupby=groupby,
            level_as_reference=True,
            nims=nims,
            final_expected_sample_size_column=final_expected_sample_size_column,
            mde_column=mde_column,
        )
        return (
            difference_df
            if verbose
            else difference_df[
                listify(groupby)
                + ["level_1", "level_2", "absolute_difference", DIFFERENCE, CI_LOWER, CI_UPPER, P_VALUE]
                + [ADJUSTED_LOWER, ADJUSTED_UPPER, ADJUSTED_P, IS_SIGNIFICANT, POWERED_EFFECT, REQUIRED_SAMPLE_SIZE]
                + ([NIM, NULL_HYPOTHESIS, PREFERENCE] if nims is not None else [])
            ]
        )

    def compute_multiple_difference(
        self,
        level: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        other_levels = [
            other for other in self._sufficient_statistics.groupby(level_columns).groups.keys() if other != level
        ]
        levels = [(level, other) for other in other_levels]
        difference_df = self._compute_differences(
            level_columns=level_columns,
            levels=levels,
            absolute=absolute,
            groupby=groupby,
            level_as_reference=level_as_reference,
            nims=nims,
            final_expected_sample_size_column=final_expected_sample_size_column,
            mde_column=mde_column,
        )
        return (
            difference_df
            if verbose
            else difference_df[
                listify(groupby)
                + [
                    "level_1",
                    "level_2",
                    "absolute_difference",
                    DIFFERENCE,
                    CI_LOWER,
                    CI_UPPER,
                    P_VALUE,
                    POWERED_EFFECT,
                    REQUIRED_SAMPLE_SIZE,
                ]
                + [ADJUSTED_LOWER, ADJUSTED_UPPER, ADJUSTED_P, IS_SIGNIFICANT]
                + ([NIM, NULL_HYPOTHESIS, PREFERENCE] if nims is not None else [])
            ]
        )

    def compute_differences(
        self,
        levels: List[Tuple],
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        difference_df = self._compute_differences(
            level_columns=level_columns,
            levels=[levels] if type(levels) == tuple else levels,
            absolute=absolute,
            groupby=groupby,
            level_as_reference=True,
            nims=nims,
            final_expected_sample_size_column=final_expected_sample_size_column,
            mde_column=mde_column,
        )
        return (
            difference_df
            if verbose
            else difference_df[
                listify(groupby)
                + ["level_1", "level_2", "absolute_difference", DIFFERENCE, CI_LOWER, CI_UPPER, P_VALUE]
                + [ADJUSTED_LOWER, ADJUSTED_UPPER, ADJUSTED_P, IS_SIGNIFICANT, POWERED_EFFECT, REQUIRED_SAMPLE_SIZE]
                + ([NIM, NULL_HYPOTHESIS, PREFERENCE] if nims is not None else [])
            ]
        )

    def _compute_differences(
        self,
        level_columns: Iterable,
        levels: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        mde_column: str,
    ):
        if type(level_as_reference) is not bool:
            raise ValueError(f"level_is_reference must be either True or False, but is {level_as_reference}.")
        groupby = listify(groupby)
        unique_levels = set([l[0] for l in levels] + [l[1] for l in levels])
        validate_levels(self._sufficient_statistics, level_columns, unique_levels)
        str2level = {level2str(lv): lv for lv in unique_levels}
        filtered_sufficient_statistics = concat(
            [self._sufficient_statistics.groupby(level_columns).get_group(group) for group in unique_levels]
        )
        levels = [
            (level2str(l[0]), level2str(l[1])) if level_as_reference else (level2str(l[1]), level2str(l[0]))
            for l in levels
        ]

        def assign_total_denominator(df, groupby):
            if self._denominator is None:
                return df.assign(**{f"current_total_{self._denominator}": None})

            if len(groupby) == 0:
                return df.assign(
                    **{f"current_total_{self._denominator}": self._sufficient_statistics[self._denominator].sum()}
                )
            else:
                return df.merge(
                    df.groupby(groupby)[self._denominator]
                    .sum()
                    .reset_index()
                    .rename(columns={self._denominator: f"current_total_{self._denominator}"})
                )

        return (
            self._sufficient_statistics.assign(
                level=self._sufficient_statistics[level_columns].agg(level2str, axis="columns")
            )
            .pipe(assign_total_denominator, groupby)
            .pipe(lambda df: df if groupby == [] else df.set_index(groupby))
            .pipe(
                self._create_comparison_df,
                groups_to_compare=levels,
                absolute=absolute,
                nims=nims,
                mde_column=mde_column,
                final_expected_sample_size_column=final_expected_sample_size_column,
                filtered_sufficient_statistics=filtered_sufficient_statistics,
            )
            .assign(level_1=lambda df: df["level_1"].map(lambda s: str2level[s]))
            .assign(level_2=lambda df: df["level_2"].map(lambda s: str2level[s]))
            .reset_index()
            .sort_values(by=groupby + ["level_1", "level_2"])
        )

    def _create_comparison_df(
        self,
        df: DataFrame,
        groups_to_compare: List[Tuple[str, str]],
        absolute: bool,
        nims: NIM_TYPE,
        mde_column: bool,
        final_expected_sample_size_column: str,
        filtered_sufficient_statistics: DataFrame,
    ) -> DataFrame:
        def join(df: DataFrame) -> DataFrame:
            has_index = not all(idx is None for idx in df.index.names)
            if has_index:
                # self-join on index (the index will typically model the date,
                # i.e., rows with the same date are joined)
                return df.merge(df, left_index=True, right_index=True, suffixes=(SFX1, SFX2))
            else:
                # join on dummy column, i.e. conduct a cross join
                return (
                    df.assign(dummy_join_column=1)
                    .merge(right=df.assign(dummy_join_column=1), on="dummy_join_column", suffixes=(SFX1, SFX2))
                    .drop(columns="dummy_join_column")
                )

        comparison_df = (
            df.pipe(add_nim_columns, nims=nims)
            .pipe(add_mde_columns, mde_column=mde_column)
            .pipe(join)
            .query(
                f"level_1 in {[l1 for l1, l2 in groups_to_compare]} and "
                + f"level_2 in {[l2 for l1, l2 in groups_to_compare]}"
                + "and level_1 != level_2"
            )
            # TODO: validate_and_rename_mdes
            .pipe(validate_and_rename_columns, NIM)
            .pipe(validate_and_rename_columns, mde_column)
            .pipe(validate_and_rename_columns, PREFERENCE)
            .pipe(validate_and_rename_columns, final_expected_sample_size_column)
            .pipe(validate_and_rename_columns, self._method_column)
            .rename(
                columns={
                    NULL_HYPOTHESIS + SFX1: NULL_HYPOTHESIS,
                    ALTERNATIVE_HYPOTHESIS + SFX1: ALTERNATIVE_HYPOTHESIS,
                    f"current_total_{self._denominator}{SFX1}": f"current_total_{self._denominator}",
                }
            )
            .drop(
                columns=[
                    NULL_HYPOTHESIS + SFX2,
                    ALTERNATIVE_HYPOTHESIS + SFX2,
                    f"current_total_{self._denominator}{SFX2}",
                ]
            )
            .assign(**{DIFFERENCE: lambda df: df[POINT_ESTIMATE + SFX2] - df[POINT_ESTIMATE + SFX1]})
            .assign(**{STD_ERR: self._std_err})
            .pipe(
                self._add_p_value_and_ci,
                final_expected_sample_size_column=final_expected_sample_size_column,
                filtered_sufficient_statistics=filtered_sufficient_statistics,
            )
            .pipe(self._add_adjusted_power)
            .apply(self._powered_effect_and_required_sample_size, mde_column=mde_column, axis=1)
            .pipe(self._adjust_if_absolute, absolute=absolute)
            .assign(**{PREFERENCE: lambda df: df[PREFERENCE].map(PREFERENCE_DICT)})
        )

        return comparison_df

    @staticmethod
    def _adjust_if_absolute(df: DataFrame, absolute: bool) -> DataFrame:
        if absolute:
            return df.assign(absolute_difference=absolute)
        else:
            return (
                df.assign(absolute_difference=absolute)
                .assign(**{DIFFERENCE: df[DIFFERENCE] / df[POINT_ESTIMATE + SFX1]})
                .assign(**{CI_LOWER: df[CI_LOWER] / df[POINT_ESTIMATE + SFX1]})
                .assign(**{CI_UPPER: df[CI_UPPER] / df[POINT_ESTIMATE + SFX1]})
                .assign(**{ADJUSTED_LOWER: df[ADJUSTED_LOWER] / df[POINT_ESTIMATE + SFX1]})
                .assign(**{ADJUSTED_UPPER: df[ADJUSTED_UPPER] / df[POINT_ESTIMATE + SFX1]})
                .assign(**{NULL_HYPOTHESIS: df[NULL_HYPOTHESIS] / df[POINT_ESTIMATE + SFX1]})
                .assign(**{POWERED_EFFECT: df[POWERED_EFFECT] / df[POINT_ESTIMATE + SFX1]})
            )

    def _corrections_power(self, number_of_success_metrics: int, number_of_guardrail_metrics: int) -> int:
        return number_of_guardrail_metrics if number_of_success_metrics == 0 else number_of_guardrail_metrics + 1

    def _add_adjusted_power(self, df: DataFrame) -> DataFrame:
        if self._correction_method in CORRECTION_METHODS_THAT_REQUIRE_METRIC_INFO:
            if self._metric_column is None or self._treatment_column is None:
                return df.assign(**{ADJUSTED_POWER: None})
            else:
                self._number_total_metrics = 1 if self._single_metric else df.groupby(self._metric_column).ngroups
                if self._single_metric:
                    if df[df[NIM].isnull()].shape[0] > 0:
                        self._number_success_metrics = 1
                    else:
                        self._number_success_metrics = 0
                else:
                    self._number_success_metrics = df[df[NIM].isnull()].groupby(self._metric_column).ngroups

                self._number_guardrail_metrics = self._number_total_metrics - self._number_success_metrics
            power_correction = self._corrections_power(
                number_of_guardrail_metrics=self._number_guardrail_metrics,
                number_of_success_metrics=self._number_success_metrics,
            )
            return df.assign(**{ADJUSTED_POWER: 1 - (1 - df[POWER]) / power_correction})
        else:
            return df.assign(**{ADJUSTED_POWER: df[POWER]})

    def _add_p_value_and_ci(
        self, df: DataFrame, final_expected_sample_size_column: str, filtered_sufficient_statistics: DataFrame
    ) -> DataFrame:
        def set_alpha_and_adjust_preference(df: DataFrame) -> DataFrame:
            alpha_0 = 1 - self._interval_size
            return (
                df.assign(
                    **{
                        ALPHA: df.apply(
                            lambda row: 2 * alpha_0
                            if self._correction_method == SPOT_1 and row[PREFERENCE] != TWO_SIDED
                            else alpha_0,
                            axis=1,
                        )
                    }
                )
                .assign(**{POWER: self._power})
                .assign(
                    **{
                        PREFERENCE_TEST: df.apply(
                            lambda row: TWO_SIDED if self._correction_method == SPOT_1 else row[PREFERENCE], axis=1
                        )
                    }
                )
            )

        def _add_adjusted_p_and_is_significant(df: DataFrame) -> DataFrame:
            if final_expected_sample_size_column is not None:
                if self._correction_method not in [
                    BONFERRONI,
                    BONFERRONI_ONLY_COUNT_TWOSIDED,
                    BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
                    SPOT_1,
                ]:
                    raise ValueError(
                        f"{self._correction_method} not supported for sequential tests. Use one of"
                        f"{BONFERRONI}, {BONFERRONI_ONLY_COUNT_TWOSIDED}, "
                        f"{BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY}, {SPOT_1}"
                    )

                groups_except_ordinal = [column for column in df.index.names if column != self._ordinal_group_column]
                n_comparisons = self._get_num_comparisons(
                    df, self._correction_method, ["level_1", "level_2"] + groups_except_ordinal
                )

                df[ADJUSTED_ALPHA] = self._compute_sequential_adjusted_alpha(
                    df, final_expected_sample_size_column, filtered_sufficient_statistics, n_comparisons
                )
                df[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE] = df[ALPHA] / n_comparisons
                df[IS_SIGNIFICANT] = df[P_VALUE] < df[ADJUSTED_ALPHA]
                df[P_VALUE] = None
                df[ADJUSTED_P] = None
            elif self._correction_method in [
                HOLM,
                HOMMEL,
                SIMES_HOCHBERG,
                SIDAK,
                HOLM_SIDAK,
                FDR_BH,
                FDR_BY,
                FDR_TSBH,
                FDR_TSBKY,
                SPOT_1_HOLM,
                SPOT_1_HOMMEL,
                SPOT_1_SIMES_HOCHBERG,
                SPOT_1_SIDAK,
                SPOT_1_HOLM_SIDAK,
                SPOT_1_FDR_BH,
                SPOT_1_FDR_BY,
                SPOT_1_FDR_TSBH,
                SPOT_1_FDR_TSBKY,
            ]:
                if self._correction_method.startswith("spot-"):
                    correction_method = self._correction_method[7:]
                else:
                    correction_method = self._correction_method

                groupby = ["level_1", "level_2"] + [column for column in df.index.names if column is not None]
                df[ADJUSTED_ALPHA] = df[ALPHA] / self._get_num_comparisons(df, self._correction_method, groupby)
                df[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE] = df[ADJUSTED_ALPHA]
                is_significant, adjusted_p, _, _ = multipletests(
                    pvals=df[P_VALUE], alpha=1 - self._interval_size, method=correction_method
                )
                df[ADJUSTED_P] = adjusted_p
                df[IS_SIGNIFICANT] = is_significant
            elif self._correction_method in [
                BONFERRONI,
                BONFERRONI_ONLY_COUNT_TWOSIDED,
                BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
                SPOT_1,
            ]:
                groupby = ["level_1", "level_2"] + [column for column in df.index.names if column is not None]
                n_comparisons = self._get_num_comparisons(df, self._correction_method, groupby)
                df[ADJUSTED_ALPHA] = df[ALPHA] / n_comparisons
                df[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE] = df[ADJUSTED_ALPHA]
                df[ADJUSTED_P] = df[P_VALUE].map(lambda p: min(p * n_comparisons, 1))
                df[IS_SIGNIFICANT] = df[P_VALUE] < df[ADJUSTED_ALPHA]
            else:
                raise ValueError("Can't figure out which correction method to use :(")

            return df

        def _add_ci(df: DataFrame) -> DataFrame:
            ci = df.apply(self._ci, axis=1, alpha_column=ALPHA)
            ci_df = DataFrame(index=ci.index, columns=[CI_LOWER, CI_UPPER], data=list(ci.values))

            if (
                self._correction_method
                in [
                    HOLM,
                    HOMMEL,
                    SIMES_HOCHBERG,
                    SPOT_1_HOLM,
                    SPOT_1_HOMMEL,
                    SPOT_1_SIMES_HOCHBERG,
                ]
                and all(df[PREFERENCE_TEST] != TWO_SIDED)
            ):
                adjusted_ci = self._ci_for_multiple_comparison_methods(
                    df,
                    correction_method=self._correction_method,
                    alpha=1 - self._interval_size,
                )
            elif self._correction_method in [
                BONFERRONI,
                BONFERRONI_ONLY_COUNT_TWOSIDED,
                BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
                SPOT_1,
                SPOT_1_HOLM,
                SPOT_1_HOMMEL,
                SPOT_1_SIMES_HOCHBERG,
                SPOT_1_SIDAK,
                SPOT_1_HOLM_SIDAK,
                SPOT_1_FDR_BH,
                SPOT_1_FDR_BY,
                SPOT_1_FDR_TSBH,
                SPOT_1_FDR_TSBKY,
            ]:
                adjusted_ci = df.apply(self._ci, axis=1, alpha_column=ADJUSTED_ALPHA)
            else:
                warn(f"Confidence intervals not supported for {self._correction_method}")
                adjusted_ci = Series(index=df.index, data=[(None, None) for i in range(len(df))])

            adjusted_ci_df = DataFrame(
                index=adjusted_ci.index, columns=[ADJUSTED_LOWER, ADJUSTED_UPPER], data=list(adjusted_ci.values)
            )

            return (
                df.assign(**{CI_LOWER: ci_df[CI_LOWER]})
                .assign(**{CI_UPPER: ci_df[CI_UPPER]})
                .assign(**{ADJUSTED_LOWER: adjusted_ci_df[ADJUSTED_LOWER]})
                .assign(**{ADJUSTED_UPPER: adjusted_ci_df[ADJUSTED_UPPER]})
            )

        return (
            df.pipe(set_alpha_and_adjust_preference)
            .assign(**{P_VALUE: lambda df: df.apply(self._p_value, axis=1)})
            .pipe(_add_adjusted_p_and_is_significant)
            .pipe(_add_ci)
        )

    def _get_num_comparisons(self, df: DataFrame, correction_method: str, groupby: Iterable) -> int:
        if correction_method == BONFERRONI:
            return max(1, df.groupby(groupby).ngroups)
        elif correction_method == BONFERRONI_ONLY_COUNT_TWOSIDED:
            return max(df.query(f'{PREFERENCE_TEST} == "{TWO_SIDED}"').groupby(groupby).ngroups, 1)
        elif correction_method in [
            HOLM,
            HOMMEL,
            SIMES_HOCHBERG,
            SIDAK,
            HOLM_SIDAK,
            FDR_BH,
            FDR_BY,
            FDR_TSBH,
            FDR_TSBKY,
        ]:
            return 1
        elif correction_method in [
            BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
            SPOT_1,
            SPOT_1_HOLM,
            SPOT_1_HOMMEL,
            SPOT_1_SIMES_HOCHBERG,
            SPOT_1_SIDAK,
            SPOT_1_HOLM_SIDAK,
            SPOT_1_FDR_BH,
            SPOT_1_FDR_BY,
            SPOT_1_FDR_TSBH,
            SPOT_1_FDR_TSBKY,
        ]:
            if self._metric_column is None or self._treatment_column is None:
                return max(1, df[df[NIM].isnull()].groupby(groupby).ngroups)
            else:
                if self._single_metric:
                    if df[df[NIM].isnull()].shape[0] > 0:
                        self._number_success_metrics = 1
                    else:
                        self._number_success_metrics = 0
                else:
                    self._number_success_metrics = df[df[NIM].isnull()].groupby(self._metric_column).ngroups

                number_comparions = len(
                    (df[self._treatment_column + SFX1] + df[self._treatment_column + SFX2]).unique()
                )
                number_segments = (
                    1
                    if len(self._segments) == 0 or not all(item in df.index.names for item in self._segments)
                    else df.groupby(self._segments).ngroups
                )

                return max(1, number_comparions * max(1, self._number_success_metrics) * number_segments)
        else:
            raise ValueError(f"Unsupported correction method: {correction_method}.")

    def achieved_power(self, level_1, level_2, mde, alpha, groupby):
        """Calculated the achieved power of test of differences between
        level 1 and level 2 given a targeted MDE.

        Args:
            level_1 (str, tuple of str): Name of first level.
            level_2 (str, tuple of str): Name of second level.
            mde (float): Absolute minimal detectable effect size.
            alpha (float): Type I error rate, cutoff value for determining
                statistical significance.
            groupby (str): Name of column.
                If specified, will return the difference for each level
                of the grouped dimension.

        Returns:
            Pandas DataFrame with the following columns:
            - level_1: Name of level 1.
            - level_2: Name of level 2.
            - power: 1 - B, where B is the likelihood of a Type II (false
                negative) error.

        """
        groupby = listify(groupby)
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        return (
            self._compute_differences(
                level_columns,
                [(level_1, level_2)],
                True,
                groupby,
                level_as_reference=True,
                nims=None,
                final_expected_sample_size_column=None,
                mde_column=None,
            )  # TODO: IS this right?
            .pipe(lambda df: df if groupby == [] else df.set_index(groupby))
            .assign(achieved_power=lambda df: df.apply(self._achieved_power, mde=mde, alpha=alpha, axis=1))
        )[["level_1", "level_2", "achieved_power"]]

    def _point_estimate(self, df: DataFrame) -> Series:
        return df.apply(lambda row: self._confidence_computers[row[self._method_column]]._point_estimate(row), axis=1)

    def _variance(self, df: DataFrame) -> Series:
        return df.apply(lambda row: self._confidence_computers[row[self._method_column]]._variance(row), axis=1)

    def _std_err(self, df: DataFrame) -> Series:
        return df.apply(lambda row: self._confidence_computers[row[self._method_column]]._std_err(row), axis=1)

    def _add_point_estimate_ci(self, df: DataFrame) -> DataFrame:
        return df.apply(
            lambda row: self._confidence_computers[row[self._method_column]]._add_point_estimate_ci(row), axis=1
        )

    def _p_value(self, row) -> float:
        if row[self._method_column] == CHI2 and row[NIM] is not None:
            raise ValueError(
                "Non-inferiority margins not supported in ChiSquared. Use StudentsTTest or ZTest instead."
            )
        return self._confidence_computers[row[self._method_column]]._p_value(row)

    def _ci(self, row, alpha_column: str) -> Tuple[float, float]:
        return self._confidence_computers[row[self._method_column]]._ci(row, alpha_column=alpha_column)

    def _powered_effect_and_required_sample_size(self, row: Series, mde_column: str) -> DataFrame:
        if row[self._method_column] != ZTEST and mde_column in row:
            raise ValueError("Minimum detectable effects only supported for ZTest.")
        elif row[self._method_column] != ZTEST or row[ADJUSTED_POWER] is None:
            row[POWERED_EFFECT] = None
            row[REQUIRED_SAMPLE_SIZE] = None
            return row
        else:
            return self._confidence_computers[row[self._method_column]]._powered_effect_and_required_sample_size(row)

    def _achieved_power(self, row: Series, mde: float, alpha: float) -> DataFrame:
        return self._confidence_computers[row[self._method_column]]._achieved_power(row, mde, alpha)

    def _compute_sequential_adjusted_alpha(
        self,
        df: DataFrame,
        final_expected_sample_size_column: str,
        filtered_sufficient_statistics: DataFrame,
        n_comparisons: int,
    ) -> Series:
        if all(df[self._method_column] == "z-test"):
            return self._confidence_computers["z-test"]._compute_sequential_adjusted_alpha(
                df, final_expected_sample_size_column, filtered_sufficient_statistics, n_comparisons
            )
        else:
            raise NotImplementedError("Sequential testing is only supported for z-tests")

    def _ci_for_multiple_comparison_methods(
        self,
        df: DataFrame,
        correction_method: str,
        alpha: float,
        w: float = 1.0,
    ) -> Tuple[Union[Series, float], Union[Series, float]]:
        if all(df[self._method_column] == "z-test"):
            return self._confidence_computers["z-test"]._ci_for_multiple_comparison_methods(
                df, correction_method, alpha, w
            )
        else:
            raise NotImplementedError(f"{self._correction_method} is only supported for ZTests")



from typing import Tuple, Union

import numpy as np
from pandas import DataFrame, Series
from scipy import optimize
from scipy import stats as st
from statsmodels.stats.weightstats import _zconfint_generic, _zstat_generic

from spotify_confidence.analysis.confidence_utils import power_calculation
from spotify_confidence.analysis.constants import (
    POINT_ESTIMATE,
    CI_LOWER,
    CI_UPPER,
    VARIANCE,
    TWO_SIDED,
    SFX2,
    SFX1,
    STD_ERR,
    PREFERENCE_TEST,
    NULL_HYPOTHESIS,
    DIFFERENCE,
    ALPHA,
    IS_SIGNIFICANT,
    HOLM,
    SPOT_1_HOLM,
    HOMMEL,
    SIMES_HOCHBERG,
    SPOT_1_HOMMEL,
    SPOT_1_SIMES_HOCHBERG,
    NIM,
    ADJUSTED_ALPHA_POWER_SAMPLE_SIZE,
    ADJUSTED_POWER,
    ALTERNATIVE_HYPOTHESIS,
    POWERED_EFFECT,
    POWERED_EFFECT_METRIC,
    REQUIRED_SAMPLE_SIZE,
    REQUIRED_SAMPLE_SIZE_METRIC,
)
from spotify_confidence.analysis.frequentist.sequential_bound_solver import bounds


def sequential_bounds(t: np.array, alpha: float, sides: int):
    return bounds(t, alpha, rho=2, ztrun=8, sides=sides, max_nints=1000)


class ZTestComputer(object):
    def __init__(self, numerator, numerator_sumsq, denominator, ordinal_group_column, interval_size):
        self._numerator = numerator
        self._numerator_sumsq = numerator_sumsq
        self._denominator = denominator
        self._ordinal_group_column = ordinal_group_column
        self._interval_size = interval_size

    def _point_estimate(self, row: Series) -> float:
        if row[self._denominator] == 0:
            raise ValueError("""Can't compute point estimate: denominator is 0""")
        return row[self._numerator] / row[self._denominator]

    def _variance(self, row: Series) -> float:
        variance = row[self._numerator_sumsq] / row[self._denominator] - row[POINT_ESTIMATE] ** 2
        if variance < 0:
            raise ValueError("Computed variance is negative. " "Please check your inputs.")
        return variance

    def _std_err(self, row: Series) -> float:
        return np.sqrt(
            row[VARIANCE + SFX1] / row[self._denominator + SFX1] + row[VARIANCE + SFX2] / row[self._denominator + SFX2]
        )

    def _add_point_estimate_ci(self, row: Series) -> Series:
        row[CI_LOWER], row[CI_UPPER] = _zconfint_generic(
            mean=row[POINT_ESTIMATE],
            std_mean=np.sqrt(row[VARIANCE] / row[self._denominator]),
            alpha=1 - self._interval_size,
            alternative=TWO_SIDED,
        )
        return row

    def _p_value(self, row: Series) -> float:
        _, p_value = _zstat_generic(
            value1=row[POINT_ESTIMATE + SFX2],
            value2=row[POINT_ESTIMATE + SFX1],
            std_diff=row[STD_ERR],
            alternative=row[PREFERENCE_TEST],
            diff=row[NULL_HYPOTHESIS],
        )
        return p_value

    def _ci(self, row: Series, alpha_column: str) -> Tuple[float, float]:
        return _zconfint_generic(
            mean=row[DIFFERENCE], std_mean=row[STD_ERR], alpha=row[alpha_column], alternative=row[PREFERENCE_TEST]
        )

    def _achieved_power(self, df: DataFrame, mde: float, alpha: float) -> DataFrame:
        v1, v2 = df[VARIANCE + SFX1], df[VARIANCE + SFX2]
        n1, n2 = df[self._denominator + SFX1], df[self._denominator + SFX2]

        var_pooled = ((n1 - 1) * v1 + (n2 - 1) * v2) / (n1 + n2 - 2)

        return power_calculation(mde, var_pooled, alpha, n1, n2)

    def _compute_sequential_adjusted_alpha(
        self,
        df: DataFrame,
        final_expected_sample_size_column: str,
        filtered_sufficient_statistics: DataFrame,
        num_comparisons: int,
    ):
        total_sample_size = (
            filtered_sufficient_statistics.groupby(df.index.names)
            .agg({self._denominator: sum, final_expected_sample_size_column: np.mean})
            .rename(columns={self._denominator: f"total_{self._denominator}"})
        )
        groups_except_ordinal = [column for column in df.index.names if column != self._ordinal_group_column]
        max_sample_size_by_group = (
            total_sample_size[f"total_{self._denominator}"].max()
            if len(groups_except_ordinal) == 0
            else total_sample_size.groupby(groups_except_ordinal)[f"total_{self._denominator}"].max()
        )

        if type(max_sample_size_by_group) is not Series:
            total_sample_size = total_sample_size.assign(
                **{f"total_{self._denominator}_max": max_sample_size_by_group}
            )
        else:
            total_sample_size = total_sample_size.merge(
                right=max_sample_size_by_group, left_index=True, right_index=True, suffixes=("", "_max")
            )

        total_sample_size = total_sample_size.assign(
            final_expected_sample_size=lambda df: df[
                [f"total_{self._denominator}_max", final_expected_sample_size_column]
            ].max(axis=1)
        ).assign(
            sample_size_proportions=lambda df: df["total_" + self._denominator] / df["final_expected_sample_size"]
        )

        def adjusted_alphas_for_group(grp: DataFrame) -> Series:
            return (
                sequential_bounds(
                    t=grp["sample_size_proportions"].values,
                    alpha=grp[ALPHA].values[0] / num_comparisons,
                    sides=2 if (grp[PREFERENCE_TEST] == TWO_SIDED).all() else 1,
                )
                .df.set_index(grp.index)
                .assign(
                    adjusted_alpha=lambda df: df.apply(
                        lambda row: 2 * (1 - st.norm.cdf(row["zb"]))
                        if (grp[PREFERENCE_TEST] == TWO_SIDED).all()
                        else 1 - st.norm.cdf(row["zb"]),
                        axis=1,
                    )
                )
            )[["zb", "adjusted_alpha"]]

        return (
            df.merge(total_sample_size, left_index=True, right_index=True)
            .groupby(groups_except_ordinal + ["level_1", "level_2"])[
                ["sample_size_proportions", PREFERENCE_TEST, ALPHA]
            ]
            .apply(adjusted_alphas_for_group)
            .reset_index()
            .set_index(df.index.names)
        )["adjusted_alpha"]

    def _ci_for_multiple_comparison_methods(
        self,
        df: DataFrame,
        correction_method: str,
        alpha: float,
        w: float = 1.0,
    ) -> Tuple[Union[Series, float], Union[Series, float]]:
        if TWO_SIDED in df[PREFERENCE_TEST]:
            raise ValueError(
                "CIs can only be produced for one-sided tests when other multiple test corrections "
                "methods than bonferroni are applied"
            )
        m_scal = len(df)
        num_significant = sum(df[IS_SIGNIFICANT])
        r = m_scal - num_significant

        def _aw(W: float, alpha: float, m_scal: float, r: int):
            return alpha * (1 - (1 - W) * (m_scal - r) / m_scal)

        def _bw(W: float, alpha: float, m_scal: float, r: int):
            return 1 - (1 - alpha) / np.power((1 - (1 - W) * (1 - np.power((1 - alpha), (1 / m_scal)))), (m_scal - r))

        if correction_method in [HOLM, SPOT_1_HOLM]:
            adjusted_alpha_rej_equal_m = 1 - alpha / m_scal
            adjusted_alpha_rej_less_m = 1 - (1 - w) * (alpha / m_scal)
            adjusted_alpha_accept = 1 - _aw(w, alpha, m_scal, r) / r if r != 0 else 0
        elif correction_method in [HOMMEL, SIMES_HOCHBERG, SPOT_1_HOMMEL, SPOT_1_SIMES_HOCHBERG]:
            adjusted_alpha_rej_equal_m = np.power((1 - alpha), (1 / m_scal))
            adjusted_alpha_rej_less_m = 1 - (1 - w) * (1 - np.power((1 - alpha), (1 / m_scal)))
            adjusted_alpha_accept = 1 - _bw(w, alpha, m_scal, r) / r if r != 0 else 0
        else:
            raise ValueError(
                "CIs not supported for correction method. "
                f"Supported methods: {HOMMEL}, {HOLM}, {SIMES_HOCHBERG},"
                f"{SPOT_1_HOLM}, {SPOT_1_HOMMEL} and {SPOT_1_SIMES_HOCHBERG}"
            )

        def _compute_ci_for_row(row: Series) -> Tuple[float, float]:
            if row[IS_SIGNIFICANT] and num_significant == m_scal:
                alpha_adj = adjusted_alpha_rej_equal_m
            elif row[IS_SIGNIFICANT] and num_significant < m_scal:
                alpha_adj = adjusted_alpha_rej_less_m
            else:
                alpha_adj = adjusted_alpha_accept

            ci_sign = -1 if row[PREFERENCE_TEST] == "larger" else 1
            bound1 = row[DIFFERENCE] + ci_sign * st.norm.ppf(alpha_adj) * row[STD_ERR]
            if ci_sign == -1:
                bound2 = max(row[NULL_HYPOTHESIS], bound1)
            else:
                bound2 = min(row[NULL_HYPOTHESIS], bound1)

            bound = bound2 if row[IS_SIGNIFICANT] else bound1

            lower = bound if row[PREFERENCE_TEST] == "larger" else -np.inf
            upper = bound if row[PREFERENCE_TEST] == "smaller" else np.inf

            return lower, upper

        return df.apply(_compute_ci_for_row, axis=1)

    def _powered_effect(
        self,
        df: Series,
        kappa: float,
        proportion_of_total: float,
        z_alpha: float,
        z_power: float,
        binary: bool,
        current_number_of_units: int,
        non_inferiority: bool,
    ) -> Series:

        if binary and not non_inferiority:
            effect = self._search_MDE_binary_local_search(
                control_avg=df[POINT_ESTIMATE + SFX1],
                control_var=df[VARIANCE + SFX1],
                non_inferiority=False,
                kappa=kappa,
                proportion_of_total=proportion_of_total,
                current_number_of_units=current_number_of_units,
                z_alpha=z_alpha,
                z_power=z_power,
            )[0]
        else:
            treatment_var = self._get_hypothetical_treatment_var(
                binary_metric=binary,
                non_inferiority=df[NIM] is not None,
                control_avg=df[POINT_ESTIMATE + SFX1],
                control_var=df[VARIANCE + SFX1],
                hypothetical_effect=0,
            )
            n2_partial = np.power((z_alpha + z_power), 2) * (df[VARIANCE + SFX1] / kappa + treatment_var)
            effect = np.sqrt((1 / (current_number_of_units * proportion_of_total)) * (n2_partial + kappa * n2_partial))

        return effect

    def _required_sample_size(
        self,
        binary: Union[Series, bool],
        non_inferiority: Union[Series, bool],
        hypothetical_effect: Union[Series, float],
        control_avg: Union[Series, float],
        control_var: Union[Series, float],
        z_alpha: float = None,
        kappa: float = None,
        proportion_of_total: Union[Series, float] = None,
        z_power: float = None,
    ) -> Union[Series, float]:

        if kappa is None:
            raise ValueError("kappa is None, must be postive float")
        if proportion_of_total is None:
            raise ValueError("proportion_of_total is None, must be between 0 and 1")

        treatment_var = np.vectorize(self._get_hypothetical_treatment_var)(
            binary, non_inferiority, control_avg, control_var, hypothetical_effect
        )

        n2 = self._treatment_group_sample_size(
            z_alpha=z_alpha,
            z_power=z_power,
            hypothetical_effect=hypothetical_effect,
            control_var=control_var,
            treatment_var=treatment_var,
            kappa=kappa,
        )
        required_sample_size = np.ceil((n2 + n2 * kappa) / proportion_of_total)
        return required_sample_size

    def _powered_effect_and_required_sample_size(
        self,
        row: Series,
    ) -> Series:
        z_alpha = st.norm.ppf(
            1 - row[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE] / (2 if row[PREFERENCE_TEST] == TWO_SIDED else 1)
        )
        z_power = st.norm.ppf(row[ADJUSTED_POWER])
        n1, n2 = row[self._denominator + SFX1], row[self._denominator + SFX2]
        kappa = n1 / n2
        binary = row[self._numerator_sumsq + SFX1] == row[self._numerator + SFX1]
        proportion_of_total = (n1 + n2) / row[f"current_total_{self._denominator}"]

        if isinstance(row[NIM], float):
            non_inferiority = not np.isnan(row[NIM])
        elif row[NIM] is None:
            non_inferiority = row[NIM] is not None
        else:
            raise ValueError("NIM has to be type float or None.")

        row[POWERED_EFFECT] = self._powered_effect(
            df=row,
            kappa=kappa,
            proportion_of_total=1,
            z_alpha=z_alpha,
            z_power=z_power,
            binary=binary,
            current_number_of_units=n1 + n2,
            non_inferiority=non_inferiority,
        )

        row[POWERED_EFFECT_METRIC] = self._powered_effect(
            df=row,
            kappa=kappa,
            proportion_of_total=proportion_of_total,
            z_alpha=z_alpha,
            z_power=z_power,
            binary=binary,
            current_number_of_units=row[f"current_total_{self._denominator}"],
            non_inferiority=non_inferiority,
        )

        if ALTERNATIVE_HYPOTHESIS in row and NULL_HYPOTHESIS in row and row[ALTERNATIVE_HYPOTHESIS] is not None:
            row[REQUIRED_SAMPLE_SIZE] = self._required_sample_size(
                proportion_of_total=1,
                z_alpha=z_alpha,
                z_power=z_power,
                binary=binary,
                non_inferiority=non_inferiority,
                hypothetical_effect=row[ALTERNATIVE_HYPOTHESIS] - row[NULL_HYPOTHESIS],
                control_avg=row[POINT_ESTIMATE + SFX1],
                control_var=row[VARIANCE + SFX1],
                kappa=kappa,
            )
            row[REQUIRED_SAMPLE_SIZE_METRIC] = self._required_sample_size(
                proportion_of_total=proportion_of_total,
                z_alpha=z_alpha,
                z_power=z_power,
                binary=binary,
                non_inferiority=non_inferiority,
                hypothetical_effect=row[ALTERNATIVE_HYPOTHESIS] - row[NULL_HYPOTHESIS],
                control_avg=row[POINT_ESTIMATE + SFX1],
                control_var=row[VARIANCE + SFX1],
                kappa=kappa,
            )
        else:
            row[REQUIRED_SAMPLE_SIZE] = None

        return row

    def _search_MDE_binary_local_search(
        self,
        control_avg: float,
        control_var: float,
        non_inferiority: bool,
        kappa: float,
        proportion_of_total: float,
        current_number_of_units: float,
        z_alpha: float = None,
        z_power: float = None,
    ):
        def f(x):
            return self.find_current_powered_effect(
                hypothetical_effect=x,
                control_avg=control_avg,
                control_var=control_var,
                binary=True,
                non_inferiority=non_inferiority,
                kappa=kappa,
                proportion_of_total=proportion_of_total,
                current_number_of_units=current_number_of_units,
                z_alpha=z_alpha,
                z_power=z_power,
            )

        max_val = 1 - control_avg
        min_val = min(10e-9, max_val)

        if min_val == max_val:
            # corner case that crashes the optimizer
            return min_val, f(min_val)

        max_iter = 100  # max number of iterations before falling back to slow grid search

        # we stop immediately if a solution was found that is "good enough". A threshold of
        # 1 indicates that
        # the approximated number of units (based on the current effect candidate) is off by
        # at most 1.0
        goodness_threshold = 1.0

        curr_iter = 0
        best_x = None
        best_fun = float("inf")

        bounds_queue = [(min_val, max_val)]

        while curr_iter < max_iter and best_fun > goodness_threshold:

            # take next value from queue
            interval = bounds_queue.pop(0)

            # conduct a bounded local search, using a very small tol value improved
            # performance during tests
            # result = optimize.minimize_scalar(f, bounds=(interval[0], interval[1]),
            # method='bounded', tol=10e-14)
            result = optimize.minimize_scalar(
                f, bounds=(interval[0], interval[1]), method="bounded", options={"xatol": 10e-14, "maxiter": 50}
            )

            if result.fun < best_fun:
                best_x = result.x
                best_fun = result.fun

            curr_iter += 1

            # add new bounds to the queue
            interval_split = (interval[0] + interval[1]) / 2
            bounds_queue.append((interval[0], interval_split))
            bounds_queue.append((interval_split, interval[1]))

        if best_fun <= goodness_threshold:
            return best_x, best_fun
        else:  # check if grid search finds a better solution
            alt_result_x, alt_result_fun = self.search_MDE_binary(
                control_avg,
                control_var,
                non_inferiority,
                kappa,
                proportion_of_total,
                current_number_of_units,
                z_alpha,
                z_power,
                return_cost_val=True,
            )

            return (alt_result_x, alt_result_fun) if alt_result_fun < best_fun else (best_x, best_fun)

    def search_MDE_binary(
        self,
        control_avg: float,
        control_var: float,
        non_inferiority: bool,
        kappa: float,
        proportion_of_total: float,
        current_number_of_units: float,
        z_alpha: float = None,
        z_power: float = None,
        return_cost_val=False,
    ):
        candidate_effects = np.linspace(10e-9, 1 - control_avg, num=2000)
        for i in range(2):
            test = []
            for effect in candidate_effects:
                test.append(
                    self.find_current_powered_effect(
                        hypothetical_effect=effect,
                        control_avg=control_avg,
                        control_var=control_var,
                        binary=True,
                        non_inferiority=non_inferiority,
                        kappa=kappa,
                        proportion_of_total=proportion_of_total,
                        current_number_of_units=current_number_of_units,
                        z_alpha=z_alpha,
                        z_power=z_power,
                    )
                )

            test = np.array(test)
            index = [idx for idx, element in enumerate(test) if element == test.min()]
            if len(index) != 1:
                index = [index[int(np.ceil(len(index) / 2))]]
            if i == 0:
                if index[0] == 9999:
                    return np.inf
                lower_effect_bound = 10e-9 if index[0] == 0 else candidate_effects[index[0] - 1]
                candidate_effects = np.linspace(lower_effect_bound, candidate_effects[index[0]], num=10000)

        index = [idx for idx, element in enumerate(test) if element == test.min()]

        return candidate_effects[index[0]], test[index[0]] if return_cost_val else candidate_effects[index[0]]

    def _treatment_group_sample_size(
        self,
        z_alpha: float,
        z_power: float,
        hypothetical_effect: float,
        control_var: float,
        treatment_var: float,
        kappa: float,
    ) -> float:
        return np.ceil(
            np.power((z_alpha + z_power) / abs(hypothetical_effect), 2) * (control_var / kappa + treatment_var)
        )

    def find_current_powered_effect(
        self,
        hypothetical_effect: float,
        control_avg: float,
        control_var: float,
        binary: bool,
        non_inferiority: bool,
        kappa: float,
        proportion_of_total: float,
        current_number_of_units: float,
        z_power: float = None,
        z_alpha: float = None,
    ) -> float:

        treatment_var = self._get_hypothetical_treatment_var(
            binary_metric=binary,
            non_inferiority=non_inferiority,
            control_avg=control_avg,
            control_var=control_var,
            hypothetical_effect=hypothetical_effect,
        )
        n2 = self._treatment_group_sample_size(
            z_alpha,
            z_power,
            hypothetical_effect,
            control_var,
            treatment_var,
            kappa,
        )

        return np.power(current_number_of_units - ((n2 + n2 * kappa) / proportion_of_total), 2)

    def _get_hypothetical_treatment_var(
        self,
        binary_metric: bool,
        non_inferiority: bool,
        control_avg: float,
        control_var: float,
        hypothetical_effect: float,
    ) -> float:
        if binary_metric and not non_inferiority:
            # For binary metrics, the variance can be derived from the average. However,
            # we do *not* do this for
            # non-inferiority tests because for non-inferiority tests, the basic assumption
            # is that the
            # mean of the control group and treatment group are identical.
            return (control_avg + hypothetical_effect) * (1 - (control_avg + hypothetical_effect))
        else:
            return control_var



from typing import Tuple

import numpy as np
from pandas import DataFrame, Series
from statsmodels.stats.proportion import proportion_confint, proportions_chisquare, confint_proportions_2indep

from spotify_confidence.analysis.confidence_utils import power_calculation
from spotify_confidence.analysis.constants import POINT_ESTIMATE, VARIANCE, CI_LOWER, CI_UPPER, SFX1, SFX2


class ChiSquaredComputer(object):
    def __init__(self, numerator, numerator_sumsq, denominator, ordinal_group_column, interval_size):
        self._numerator = numerator
        self._numerator_sumsq = numerator_sumsq
        self._denominator = denominator
        self._ordinal_group_column = ordinal_group_column
        self._interval_size = interval_size

    def _point_estimate(self, row: Series) -> float:
        if row[self._denominator] == 0:
            raise ValueError("""Can't compute point estimate: denominator is 0""")
        return row[self._numerator] / row[self._denominator]

    def _variance(self, row: Series) -> float:
        variance = row[POINT_ESTIMATE] * (1 - row[POINT_ESTIMATE])
        if variance < 0:
            raise ValueError("Computed variance is negative. " "Please check your inputs.")
        return variance

    def _std_err(self, row: Series) -> float:
        return np.sqrt(
            row[VARIANCE + SFX1] / row[self._denominator + SFX1] + row[VARIANCE + SFX2] / row[self._denominator + SFX2]
        )

    def _add_point_estimate_ci(self, row: DataFrame) -> Series:
        row[CI_LOWER], row[CI_UPPER] = proportion_confint(
            count=row[self._numerator],
            nobs=row[self._denominator],
            alpha=1 - self._interval_size,
        )
        return row

    def _p_value(self, row: Series) -> float:
        _, p_value, _ = proportions_chisquare(
            count=[row[self._numerator + SFX1], row[self._numerator + SFX2]],
            nobs=[row[self._denominator + SFX1], row[self._denominator + SFX2]],
        )
        return p_value

    def _ci(self, row: Series, alpha_column: str) -> Tuple[float, float]:
        return confint_proportions_2indep(
            count1=row[self._numerator + SFX2],
            nobs1=row[self._denominator + SFX2],
            count2=row[self._numerator + SFX1],
            nobs2=row[self._denominator + SFX1],
            alpha=row[alpha_column],
            compare="diff",
            method="wald",
        )

    def _achieved_power(self, df: DataFrame, mde: float, alpha: float) -> DataFrame:
        s1, s2 = df[self._numerator + SFX1], df[self._numerator + SFX2]
        n1, n2 = df[self._denominator + SFX1], df[self._denominator + SFX2]

        pooled_prop = (s1 + s2) / (n1 + n2)
        var_pooled = pooled_prop * (1 - pooled_prop)

        return power_calculation(mde, var_pooled, alpha, n1, n2)



"""Tests for `confidence` categorical variables."""

import pytest
import spotify_confidence
import pandas as pd
import numpy as np
from spotify_confidence.analysis.frequentist.confidence_computers.t_test_computer import TTestComputer
from spotify_confidence.analysis.frequentist.confidence_computers.generic_computer import GenericComputer
from spotify_confidence.analysis.constants import (
    POINT_ESTIMATE,
    VARIANCE,
    SFX1,
    SFX2,
    INCREASE_PREFFERED,
    DECREASE_PREFFERED,
    DIFFERENCE,
    P_VALUE,
    CI_UPPER,
    CI_LOWER,
)


def chart_data(chart_object, series_name):
    """Retrieve data from chart object based on series name.

    Note: if there's only one series the name is None.

    Returns:
        Dictionary populated with data from the chart.
    """
    cannonical_series_name = chart_object.plot._cannonical_series_name(series_name)
    return chart_object.figure.select(cannonical_series_name)[0].data


class TestCategorical(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "nr_of_items": [1969, 312, 2955, 195, 24, 330],
                "nr_of_items_sumsq": [5767, 984, 8771, 553, 80, 1010],
                "users": [1009, 104, 1502, 100, 10, 150],
                "country": [
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["country", "variation_name"],
            interval_size=0.95,
        )

    def test_sample_variance(self):
        n = 1000
        x = np.random.randn(n)
        var = x.var()

        sum_squared = np.sum(x * x)

        comp = TTestComputer(
            numerator=None, numerator_sumsq="sumsq", denominator="n", ordinal_group_column=None, interval_size=None
        )
        var_to_verify = pd.DataFrame({"sumsq": [sum_squared], POINT_ESTIMATE: [x.mean()], "n": [n]}).apply(
            lambda row: comp._variance(row), axis=1
        )

        assert np.allclose(var_to_verify, var)

    def test_standard_error_diff_mean_from_sums(self):
        n1 = int(1e6)
        mu1 = 1
        sigma1 = 1
        x1 = np.random.normal(mu1, sigma1, n1)

        n2 = int(1e6)
        mu2 = 2
        sigma2 = 2
        x2 = np.random.normal(mu2, sigma2, n2)

        std_diff = np.sqrt(sigma1 ** 2 / n1 + sigma2 ** 2 / n2)

        comp = GenericComputer(
            data_frame=pd.DataFrame({"group": ["control"], "success": [1], "n": [2], "method": ["t-test"]}),
            numerator_column="success",
            numerator_sum_squares_column="sumsq",
            denominator_column="n",
            categorical_group_columns=["group"],
            ordinal_group_column=None,
            metric_column=None,
            treatment_column=None,
            interval_size=None,
            power=0.8,
            correction_method="Bonferroni",
            method_column="method",
            bootstrap_samples_column=None,
        )
        diff_se = comp._std_err(
            pd.DataFrame(
                {
                    VARIANCE + SFX1: [x1.var()],
                    VARIANCE + SFX2: [x2.var()],
                    "n" + SFX1: [n1],
                    "n" + SFX2: [n2],
                    "method": ["t-test"],
                }
            )
        )

        assert np.allclose(std_diff, diff_se, atol=1e-5)

    def test_summary(self):
        summary = self.test.summary()

        assert np.array_equal(summary.country, np.array(["us", "us", "us", "gb", "gb", "gb"]))
        assert np.array_equal(summary.point_estimate, self.data.nr_of_items / self.data.users)

        assert np.allclose(summary["ci_lower"], np.array([1.866117, 2.867880, 1.896356, 1.689206, 1.329352, 1.977998]))

        assert np.allclose(summary["ci_upper"], np.array([2.036757, 3.132120, 2.038397, 2.210794, 3.470648, 2.422002]))

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1
        ch = self.test.summary_plot(groupby="country")
        assert len(ch.charts) == 2

    def test_p_value_is_symmetric_and_correct(self):
        df = pd.DataFrame(
            {
                "group_name": ["Control", "Test"],
                "users": [48351, 50571],
                "nr_of_items": [1.438602e06, 1.521974e06],
                "nr_of_items_sumsq": [7.330581e07, 7.862121e07],
            }
        )

        ttest = spotify_confidence.StudentsTTest(
            data_frame=df,
            denominator_column="users",
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            categorical_group_columns="group_name",
        )

        diff_summary_1 = ttest.difference(level_1="Control", level_2="Test")
        diff_summary_2 = ttest.difference(level_1="Test", level_2="Control")

        assert diff_summary_1["p-value"].iloc[0] == diff_summary_2["p-value"].iloc[0]
        assert np.isclose(diff_summary_1["p-value"].iloc[0], 0.03334, 0.01)

    def test_difference(self):
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], -1.0485629)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.205402)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.891723)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000)

        diff = self.test.difference(("gb", "control"), ("gb", "test2"))
        assert np.allclose(diff["difference"].iloc[0], -0.2)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.283253)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.883253)
        assert np.allclose(diff["p-value"].iloc[0], 0.689599)

        diff = self.test.difference(("us", "control"), ("us", "test"), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], -0.349520)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.4018008)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.297241)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000)

        diff = self.test.difference("control", "test", groupby="country")
        assert np.allclose(diff["difference"], np.array([-0.450000, -1.048563]))
        assert np.allclose(diff["ci_lower"], np.array([-1.538289, -1.205402]))
        assert np.allclose(diff["ci_upper"], np.array([0.638289, -0.891723]))
        assert np.allclose(diff["p-value"], np.array([0.380282, 0.000000]))

    def test_difference_with_interval_sizes(self):
        """
        https: // www.quantitativeskills.com / sisa / statistics / t - test.htm
        was used to validate results
        """
        self.test._confidence_computer._interval_size = 0.99

        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], -1.0485629)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.255391)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.841735)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000)

        self.test._confidence_computer._interval_size = 0.999
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], -1.0485629)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.314140)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.782986)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000)

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("bad_value", "bad_value"), ("bad_value", "bad_value"))

        ch = self.test.difference_plot(("us", "control"), ("us", "test"))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="country")
        assert len(ch.charts) == 1

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(("bad_value", "bad_value"), level_as_reference=False)

        diff = self.test.multiple_difference(("us", "control"), level_as_reference=False)
        assert len(diff) > 0

        diff = self.test.multiple_difference("test", groupby="country", level_as_reference=False)
        assert len(diff) > 0

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(
                ("bad_value", "bad_value"), ("bad_value", "bad_value"), level_as_reference=False
            )

        ch = self.test.multiple_difference_plot(("us", "control"), level_as_reference=False)
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot("test", groupby="country", level_as_reference=False)
        assert len(ch.charts) == 1

    def test_achieved_power(self):
        power_df = self.test.achieved_power(level_1=("us", "test2"), level_2=("us", "test"), mde=0.2, alpha=0.05)
        power = power_df.achieved_power
        expected_power = 0.9409124  # Based on G*Power calculation

        assert np.allclose(power, expected_power, 0.005)


class TestOrdinal(object):
    def setup(self):

        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [500, 8, 100, 510, 8, 100, 520, 9, 104, 530, 7, 100, 530, 8, 103],
                "nr_of_items_sumsq": [2500, 12, 150, 2510, 13, 140, 2520, 14, 154, 2530, 15, 160, 2530, 16, 103],
                "users": [1010, 22, 150, 1000, 20, 153, 1030, 23, 154, 1000, 20, 150, 1040, 21, 155],
                "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns="variation_name",
            ordinal_group_column="days_since_reg",
            interval_size=0.95,
        )

    def test_summary(self):
        summary = self.test.summary()

        assert np.array_equal(summary.point_estimate, self.data.nr_of_items / self.data.users)

        assert np.allclose(
            summary["ci_lower"],
            np.array(
                [
                    0.402840,
                    0.078624,
                    0.546410,
                    0.416920,
                    0.072390,
                    0.542033,
                    0.414337,
                    0.099428,
                    0.557913,
                    0.436937,
                    -0.020737,
                    0.539399,
                    0.419921,
                    0.023464,
                    0.589596,
                ]
            ),
        )

        assert np.allclose(
            summary["ci_upper"],
            np.array(
                [
                    0.587260,
                    0.648649,
                    0.786923,
                    0.603080,
                    0.727610,
                    0.765157,
                    0.595372,
                    0.683181,
                    0.792736,
                    0.623063,
                    0.720737,
                    0.793934,
                    0.599310,
                    0.738441,
                    0.739436,
                ]
            ),
        )

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

        ch = self.test.summary_plot().charts[0]
        np.array_equal(chart_data(ch, "control")["days_since_reg"], np.array([1, 2, 3, 4, 5, 5, 4, 3, 2, 1]))

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"))

        diff = self.test.difference(("control", 1), ("test", 1))
        assert np.allclose(diff["difference"].iloc[0], 0.131413)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.166276)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.429102)
        assert np.allclose(diff["p-value"].iloc[0], 0.372650)

        diff = self.test.difference(("control", 1), ("test", 1), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], 0.361386)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.457258)
        assert np.allclose(diff["ci_upper"].iloc[0], 1.1800302)
        assert np.allclose(diff["p-value"].iloc[0], 0.37265075)

        diff = self.test.difference("control", "test", groupby="days_since_reg")
        assert np.allclose(diff["difference"], np.array([0.13141314, 0.11, 0.11355002, 0.18, 0.128663]))

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("control", 10), ("test", 10))

        ch = self.test.difference_plot(("control", 1), ("test", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="days_since_reg")
        assert len(ch.charts) == 1

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(
                ("bad_value", "bad_value"), ("bad_value", "bad_value"), level_as_reference=False
            )

        diff = self.test.multiple_difference(("control", 1), level_as_reference=False)
        assert len(diff) > 0

        diff = self.test.multiple_difference("test", groupby="days_since_reg", level_as_reference=False)
        assert len(diff) > 0

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(
                ("bad_value", "bad_value"), ("bad_value", "bad_value"), level_as_reference=False
            )

        ch = self.test.multiple_difference_plot(("control", 1), level_as_reference=False)
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot("test", groupby="days_since_reg", level_as_reference=False)
        assert len(ch.charts) == 1

    def test_achieved_power(self):
        power_df = self.test.achieved_power(level_1=("control", 1), level_2=("test", 1), mde=1, alpha=0.05)
        power = power_df.achieved_power
        expected_power = 0.8790801  # Based on G*Power calculation
        assert np.allclose(power, expected_power, 0.005)


class TestOrdinalPlusTwoCategorical(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                ],
                "nr_of_items_sumsq": [
                    1010,
                    32,
                    250,
                    1000,
                    30,
                    253,
                    1030,
                    33,
                    254,
                    1000,
                    30,
                    250,
                    1040,
                    31,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "users": [
                    2010,
                    42,
                    250,
                    2000,
                    40,
                    253,
                    2030,
                    43,
                    254,
                    2000,
                    40,
                    250,
                    2040,
                    41,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "days_since_reg": [
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                ],
                "country": [
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "country"],
            ordinal_group_column="days_since_reg",
        )

    def test_summary_plot(self):
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

        ch = self.test.summary_plot(groupby=["country"])
        assert len(ch.charts) == 2

        ch = self.test.summary_plot(groupby=["days_since_reg"])
        assert len(ch.charts) == 5

        ch = self.test.summary_plot(groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 10

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"), groupby="days_since_reg")

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby="days_since_reg")
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby=["days_since_reg"])
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", 1), level_2=("control", 1), groupby=["country"])
        assert len(df) == 2
        assert "country" in df.columns

    def test_differece_with_nims(self):
        df = self.test.difference(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby="days_since_reg",
            non_inferiority_margins=(0.01, INCREASE_PREFFERED),
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby=["days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", 1),
            level_2=("control", 1),
            groupby=["country"],
            non_inferiority_margins={"us": (0.01, INCREASE_PREFFERED), "gb": (0.05, INCREASE_PREFFERED)},
        )
        assert len(df) == 2
        assert "country" in df.columns

        df = self.test.difference(
            level_1="test",
            level_2="control",
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        nims = {
            ("us", 1): (0.01, DECREASE_PREFFERED),
            ("us", 2): (0.1, INCREASE_PREFFERED),
            ("us", 3): (0.2, DECREASE_PREFFERED),
            ("us", 4): (0.5, INCREASE_PREFFERED),
            ("us", 5): (0.99, DECREASE_PREFFERED),
            ("gb", 1): (1.01, INCREASE_PREFFERED),
            ("gb", 2): (2.01, DECREASE_PREFFERED),
            ("gb", 3): (3.01, INCREASE_PREFFERED),
            ("gb", 4): (4.01, DECREASE_PREFFERED),
            ("gb", 5): (5.01, INCREASE_PREFFERED),
        }
        df = self.test.difference(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=nims
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        df = self.test.multiple_difference(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=nims,
        )
        assert len(df) == 20

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("control", "us", 10), ("test", "us", 10))

        ch = self.test.difference_plot(level_1=("control", "us", 1), level_2=("test", "us", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", "us"), level_2=("test", "us"), groupby="days_since_reg")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1="control", level_2="test", groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", 1), level_2=("test", 1), groupby="country")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", "us"), level_2=("test", "us"), groupby="days_since_reg")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1="control", level_2="test", groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 1

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(level="control")

        ch = self.test.multiple_difference_plot(
            level="control", groupby=["country", "days_since_reg"], level_as_reference=True
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level="us", groupby=["variation_name", "days_since_reg"], level_as_reference=True
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level=1, groupby=["country", "variation_name"], level_as_reference=True
        )
        assert len(ch.charts) == 1

    def test_differece_plot_with_nims(self):
        ch = self.test.difference_plot(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby="days_since_reg",
            non_inferiority_margins=(0.01, INCREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby=["days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", 1),
            level_2=("control", 1),
            groupby=["country"],
            non_inferiority_margins={"us": (0.01, INCREASE_PREFFERED), "gb": (0.05, INCREASE_PREFFERED)},
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1="test",
            level_2="control",
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        nims = {
            ("us", 1): (0.01, DECREASE_PREFFERED),
            ("us", 2): (0.1, INCREASE_PREFFERED),
            ("us", 3): (0.2, DECREASE_PREFFERED),
            ("us", 4): (0.5, INCREASE_PREFFERED),
            ("us", 5): (0.99, DECREASE_PREFFERED),
            ("gb", 1): (1.01, INCREASE_PREFFERED),
            ("gb", 2): (2.01, DECREASE_PREFFERED),
            ("gb", 3): (3.01, INCREASE_PREFFERED),
            ("gb", 4): (4.01, DECREASE_PREFFERED),
            ("gb", 5): (5.01, INCREASE_PREFFERED),
        }
        ch = self.test.difference_plot(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=nims
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=nims,
        )
        assert len(ch.charts) == 1


class TestCategoricalBinomialData(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "success": [500, 42, 1005, 50, 4, 100],
                "total": [1009, 104, 1502, 100, 10, 150],
                "country": [
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column="success",
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
        )

    def test_summary(self):
        """Area plot tests"""

        summary = self.test.summary()
        assert np.array_equal(summary.country, np.array(["us", "us", "us", "gb", "gb", "gb"]))
        assert np.array_equal(summary.point_estimate, self.data.success / self.data.total)
        assert np.allclose(summary["ci_lower"], np.array([0.464653, 0.308424, 0.645293, 0.400789, 0.049548, 0.590610]))
        assert np.allclose(summary["ci_upper"], np.array([0.526427, 0.499269, 0.692923, 0.599211, 0.750452, 0.742723]))

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(("bad_value", "bad_value"), level_as_reference=False)

        diff = self.test.multiple_difference(("us", "control"), level_as_reference=False)
        assert np.allclose(
            diff["adjusted p-value"], np.array([1e00, 8.36850129e-01, 1.49172324e-04, 3.62369184e-01, 2.27154393e-06])
        )
        assert np.allclose(
            diff["p-value"], np.array([9.81516149e-01, 1.67370026e-01, 2.98344648e-05, 7.24738369e-02, 4.54308787e-07])
        )
        assert np.allclose(
            diff["adjusted ci_lower"], np.array([-0.501747, -0.276600, -0.422946, -0.224093, -0.395207])
        )
        assert np.allclose(diff["adjusted ci_upper"], np.array([0.509440, 0.084292, -0.102695, 0.040705, -0.135317]))

        diff = self.test.multiple_difference("test", groupby="country", level_as_reference=False)
        assert np.allclose(diff["adjusted p-value"], np.array([1, 0.035594, 0.289895, 0.0]))
        assert np.allclose(diff["p-value"], np.array([0.55155672, 0.00889848, 0.07247384, 0.0]))
        assert np.allclose(diff["adjusted ci_lower"], np.array([-0.385571, -0.325682, -0.036587, -0.223262]))
        assert np.allclose(diff["adjusted ci_upper"], np.array([0.585571, -0.0076513, 0.219975, -0.123874]))


class TestWithNims(object):
    def setup(self):
        self.data = pd.DataFrame(
            [
                {
                    "group": "1",
                    "count": 5000,
                    "sum": 10021.0,
                    "sum_of_squares": 25142.0,
                    "avg": 2.004210,
                    "var": 1.0116668,
                },
                {
                    "group": "2",
                    "count": 5000,
                    "sum": 9892.0,
                    "sum_of_squares": 24510.0,
                    "avg": 1.978424,
                    "var": 0.9881132,
                },
            ]
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="sum",
            numerator_sum_squares_column="sum_of_squares",
            denominator_column="count",
            categorical_group_columns="group",
            interval_size=0.99,
        )

    def test_compare_series_non_inferiority_improve_postitive(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "increase"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_UPPER].values[0] == float("inf")
        np.testing.assert_almost_equal(diff[CI_LOWER].values[0], -0.0723, 3)
        assert diff[P_VALUE].values[0] > 0.01

    def test_compare_series_non_inferiority_improve_negative(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "decrease"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_LOWER].values[0] == -float("inf")
        np.testing.assert_almost_equal(diff[CI_UPPER].values[0], 0.0207, 3)
        assert diff[P_VALUE].values[0] < 0.01

    def test_one_sided_ztest_positive(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(None, "increase"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_UPPER].values[0] == float("inf")
        np.testing.assert_almost_equal(diff[CI_LOWER].values[0], -0.0723, 3)
        assert diff[P_VALUE].values[0] > 0.01

    def test_one_sided_ztest_negative(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(None, "decrease"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_LOWER].values[0] == -float("inf")
        np.testing.assert_almost_equal(diff[CI_UPPER].values[0], 0.0207, 3)
        assert diff[P_VALUE].values[0] > 0.01
